# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining GraphBuilder class and related methods.
"""

from __future__ import annotations

__all__ = ["GraphBuilder"]


from collections.abc import Iterable, Mapping
from typing import Any

from lsst.daf.butler import Butler, DataCoordinate, Datastore, Registry
from lsst.daf.butler.registry.wildcards import CollectionWildcard

from ._datasetQueryConstraints import DatasetQueryConstraintVariant
from .all_dimensions_quantum_graph_builder import AllDimensionsQuantumGraphBuilder
from .graph import QuantumGraph
from .pipeline import Pipeline, TaskDef
from .pipeline_graph import PipelineGraph

# Re-exports for backwards-compatibility.
from .quantum_graph_builder import GraphBuilderError  # noqa: F401
from .quantum_graph_builder import OutputExistsError  # noqa: F401
from .quantum_graph_builder import PrerequisiteMissingError  # noqa: F401


class GraphBuilder:
    """GraphBuilder class is responsible for building task execution graph from
    a Pipeline.

    Parameters
    ----------
    registry : `~lsst.daf.butler.Registry`
        Data butler instance.
    skipExistingIn : `~typing.Any`
        Expressions representing the collections to search for existing
        output datasets that should be skipped.  See
        :ref:`daf_butler_ordered_collection_searches`.
    clobberOutputs : `bool`, optional
        If `True` (default), allow quanta to created even if partial outputs
        exist; this requires the same behavior behavior to be enabled when
        executing.
    datastore : `~lsst.daf.butler.Datastore`, optional
        If not `None` then fill datastore records in each generated Quantum.
    """

    def __init__(
        self,
        registry: Registry,
        skipExistingIn: Any = None,
        clobberOutputs: bool = True,
        datastore: Datastore | None = None,
    ):
        self.registry = registry
        self.dimensions = registry.dimensions
        self.skipExistingIn = skipExistingIn
        self.clobberOutputs = clobberOutputs
        self.datastore = datastore

    def makeGraph(
        self,
        pipeline: Pipeline | Iterable[TaskDef],
        collections: Any,
        run: str,
        userQuery: str | None,
        datasetQueryConstraint: DatasetQueryConstraintVariant = DatasetQueryConstraintVariant.ALL,
        metadata: Mapping[str, Any] | None = None,
        bind: Mapping[str, Any] | None = None,
        dataId: DataCoordinate | None = None,
    ) -> QuantumGraph:
        """Create execution graph for a pipeline.

        Parameters
        ----------
        pipeline : `Pipeline` or `~collections.abc.Iterable` [ `TaskDef` ]
            Pipeline definition, task names/classes and their configs.
        collections : `~typing.Any`
            Expressions representing the collections to search for input
            datasets.  See :ref:`daf_butler_ordered_collection_searches`.
        run : `str`
            Name of the `~lsst.daf.butler.CollectionType.RUN` collection for
            output datasets. Collection does not have to exist and it will be
            created when graph is executed.
        userQuery : `str`
            String which defines user-defined selection for registry, should be
            empty or `None` if there is no restrictions on data selection.
        datasetQueryConstraint : `DatasetQueryConstraintVariant`, optional
            The query constraint variant that should be used to constraint the
            query based on dataset existance, defaults to
            `DatasetQueryConstraintVariant.ALL`.
        metadata : Optional Mapping of `str` to primitives
            This is an optional parameter of extra data to carry with the
            graph.  Entries in this mapping should be able to be serialized in
            JSON.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``userQuery`` expression, keyed by the identifiers they replace.
        dataId : `lsst.daf.butler.DataCoordinate`, optional
            Data ID that should also be included in the query constraint.
            Ignored if ``pipeline`` is a `Pipeline` instance (which has its own
            data ID).

        Returns
        -------
        graph : `QuantumGraph`
            The constructed graph.

        Raises
        ------
        UserExpressionError
            Raised when user expression cannot be parsed.
        OutputExistsError
            Raised when output datasets already exist.
        Exception
            Other exceptions types may be raised by underlying registry
            classes.
        """
        if isinstance(pipeline, Pipeline):
            pipeline_graph = pipeline.to_graph()
        else:
            pipeline_graph = PipelineGraph(data_id=dataId)
            for task_def in pipeline:
                pipeline_graph.add_task(
                    task_def.label,
                    task_def.taskClass,
                    config=task_def.config,
                    connections=task_def.connections,
                )
        # We assume `registry` is actually a RegistryShim that has a butler
        # inside it, since that's now the only kind of Registry code outside
        # Butler should be able to get, and we assert that the datastore came
        # from the same place.  Soon this interface will be deprecated in favor
        # of QuantumGraphBuilder (which takes a Butler directly, as all new
        # code should) anyway.
        butler: Butler = self.registry._butler  # type: ignore
        assert butler._datastore is self.datastore or self.datastore is None
        qgb = AllDimensionsQuantumGraphBuilder(
            pipeline_graph,
            butler,
            input_collections=CollectionWildcard.from_expression(collections).require_ordered(),
            output_run=run,
            skip_existing_in=self.skipExistingIn if self.skipExistingIn is not None else (),
            clobber=self.clobberOutputs,
            where=userQuery if userQuery is not None else "",
            dataset_query_constraint=datasetQueryConstraint,
            bind=bind,
        )
        return qgb.build(metadata)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping, Sequence
from typing import Any, ClassVar, TypeVar, cast, overload

import networkx

from ._dataset_types import DatasetTypeNode
from ._exceptions import UnresolvedGraphError
from ._nodes import NodeKey, NodeType
from ._tasks import TaskInitNode, TaskNode

_N = TypeVar("_N", covariant=True)
_T = TypeVar("_T")


class MappingView(Mapping[str, _N]):
    """Base class for mapping views into nodes of certain types in a
    `PipelineGraph`.

    Parameters
    ----------
    parent_xgraph : `networkx.MultiDiGraph`
        Backing networkx graph for the `PipelineGraph` instance.

    Notes
    -----
    Instances should only be constructed by `PipelineGraph` and its helper
    classes.

    Iteration order is topologically sorted if and only if the backing
    `PipelineGraph` has been sorted since its last modification.
    """

    def __init__(self, parent_xgraph: networkx.MultiDiGraph) -> None:
        self._parent_xgraph = parent_xgraph
        self._keys: list[str] | None = None

    _NODE_TYPE: ClassVar[NodeType]  # defined by derived classes

    def __contains__(self, key: object) -> bool:
        # The given key may not be a str, but if it isn't it'll just fail the
        # check, which is what we want anyway.
        return NodeKey(self._NODE_TYPE, cast(str, key)) in self._parent_xgraph

    def __iter__(self) -> Iterator[str]:
        if self._keys is None:
            self._keys = self._make_keys(self._parent_xgraph)
        return iter(self._keys)

    def __getitem__(self, key: str) -> _N:
        return self._parent_xgraph.nodes[NodeKey(self._NODE_TYPE, key)]["instance"]

    def __len__(self) -> int:
        if self._keys is None:
            self._keys = self._make_keys(self._parent_xgraph)
        return len(self._keys)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self!s})"

    def __str__(self) -> str:
        return f"{{{', '.join(iter(self))}}}"

    def _reorder(self, parent_keys: Sequence[NodeKey]) -> None:
        """Set this view's iteration order according to the given iterable of
        parent keys.

        Parameters
        ----------
        parent_keys : `~collections.abc.Sequence` [ `NodeKey` ]
            Superset of the keys in this view, in the new order.
        """
        self._keys = self._make_keys(parent_keys)

    def _reset(self) -> None:
        """Reset all cached content.

        This should be called by the parent graph after any changes that could
        invalidate the view, causing it to be reconstructed when next
        requested.
        """
        self._keys = None

    def _make_keys(self, parent_keys: Iterable[NodeKey]) -> list[str]:
        """Make a sequence of keys for this view from an iterable of parent
        keys.

        Parameters
        ----------
        parent_keys : `~collections.abc.Iterable` [ `NodeKey` ]
            Superset of the keys in this view.
        """
        return [str(k) for k in parent_keys if k.node_type is self._NODE_TYPE]


class TaskMappingView(MappingView[TaskNode]):
    """A mapping view of the tasks in a `PipelineGraph`.

    Notes
    -----
    Mapping keys are task labels and values are `TaskNode` instances.
    Iteration order is topological if and only if the `PipelineGraph` has been
    sorted since its last modification.
    """

    _NODE_TYPE = NodeType.TASK


class TaskInitMappingView(MappingView[TaskInitNode]):
    """A mapping view of the nodes representing task initialization in a
    `PipelineGraph`.

    Notes
    -----
    Mapping keys are task labels and values are `TaskInitNode` instances.
    Iteration order is topological if and only if the `PipelineGraph` has been
    sorted since its last modification.
    """

    _NODE_TYPE = NodeType.TASK_INIT


class DatasetTypeMappingView(MappingView[DatasetTypeNode]):
    """A mapping view of the nodes representing task initialization in a
    `PipelineGraph`.

    Notes
    -----
    Mapping keys are parent dataset type names and values are `DatasetTypeNode`
    instances, but values are only available for nodes that have been resolved
    (see `PipelineGraph.resolve`).  Attempting to access an unresolved value
    will result in `UnresolvedGraphError` being raised.  Keys for unresolved
    nodes are always present and iterable.

    Iteration order is topological if and only if the `PipelineGraph` has been
    sorted since its last modification.
    """

    _NODE_TYPE = NodeType.DATASET_TYPE

    def __getitem__(self, key: str) -> DatasetTypeNode:
        if (result := super().__getitem__(key)) is None:
            raise UnresolvedGraphError(f"Node for dataset type {key!r} has not been resolved.")
        return result

    def is_resolved(self, key: str) -> bool:
        """Test whether a node has been resolved.

        Parameters
        ----------
        key : `str`
            Node to check.

        Returns
        -------
        `bool`
            Whether the node has been resolved or not.
        """
        return super().__getitem__(key) is not None

    @overload
    def get_if_resolved(self, key: str) -> DatasetTypeNode | None: ...  # pragma: nocover

    @overload
    def get_if_resolved(self, key: str, default: _T) -> DatasetTypeNode | _T: ...  # pragma: nocover

    def get_if_resolved(self, key: str, default: Any = None) -> DatasetTypeNode | Any:
        """Get a node or return a default if it has not been resolved.

        Parameters
        ----------
        key : `str`
            Parent dataset type name.
        default : `~typing.Any`
            Value to return if this dataset type has not been resolved.

        Returns
        -------
        result : `DatasetTypeNode`
            The resolved node or the default value.

        Raises
        ------
        KeyError
            Raised if the node is not present in the graph at all.
        """
        if (result := super().__getitem__(key)) is None:
            return default  # type: ignore
        return result



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("PexConfigFormatter",)

import os.path
from typing import Any

from lsst.daf.butler.formatters.file import FileFormatter
from lsst.pex.config import Config


class PexConfigFormatter(FileFormatter):
    """Formatter implementation for reading and writing
    `lsst.pex.config.Config` instances.
    """

    extension = ".py"

    def _readFile(self, path: str, pytype: type[Any] | None = None) -> Any:
        """Read a pex.config.Config instance from the given file.

        Parameters
        ----------
        path : `str`
            Path to use to open the file.
        pytype : `type`, optional
            Class to use to read the config file.

        Returns
        -------
        data : `lsst.pex.config.Config`
            Instance of class ``pytype`` read from config file. `None`
            if the file could not be opened.
        """
        if not os.path.exists(path):
            return None

        # Automatically determine the Config class from the serialized form
        with open(path, "r") as fd:
            config_py = fd.read()
        return Config._fromPython(config_py)

    def _writeFile(self, inMemoryDataset: Any) -> None:
        """Write the in memory dataset to file on disk.

        Parameters
        ----------
        inMemoryDataset : `object`
            Object to serialize.

        Raises
        ------
        Exception
            The file could not be written.
        """
        inMemoryDataset.save(self.fileDescriptor.location.path)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any

import click
from lsst.daf.butler.cli.opt import (
    options_file_option,
    register_dataset_types_option,
    repo_argument,
    transfer_dimensions_option,
)
from lsst.daf.butler.cli.utils import ButlerCommand

from ... import script
from ..opt import instrument_argument, update_output_chain_option


@click.command(short_help="Add an instrument definition to the repository", cls=ButlerCommand)
@repo_argument(required=True)
@instrument_argument(required=True, nargs=-1, help="The fully-qualified name of an Instrument subclass.")
@click.option("--update", is_flag=True)
def register_instrument(*args: Any, **kwargs: Any) -> None:
    """Add an instrument to the data repository."""
    script.register_instrument(*args, **kwargs)


@click.command(short_help="Transfer datasets from a graph to a butler.", cls=ButlerCommand)
@click.argument("graph", required=True)
@click.argument("dest", required=True)
@register_dataset_types_option()
@transfer_dimensions_option(default=False)
@update_output_chain_option()
@click.option(
    "--dry-run", is_flag=True, default=False, help="Run the transfer but do not update the destination butler"
)
@options_file_option()
def transfer_from_graph(**kwargs: Any) -> None:
    """Transfer datasets from a quantum graph to a destination butler.

    SOURCE is a URI to the Butler repository containing the RUN dataset.

    DEST is a URI to the Butler repository that will receive copies of the
    datasets.
    """
    number = script.transfer_from_graph(**kwargs)
    print(f"Number of datasets transferred: {number}")



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
import textwrap
import unittest

import lsst.utils.tests
from lsst.pipe.base.pipelineIR import ConfigIR, PipelineIR, PipelineSubsetCtrl

# Find where the test pipelines exist and store it in an environment variable.
os.environ["TESTDIR"] = os.path.dirname(__file__)


class ConfigIRTestCase(unittest.TestCase):
    """A test case for ConfigIR Objects.

    ConfigIR contains a method that is not exercised by the PipelineIR task,
    so it should be tested here.
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testMergeConfig(self):
        # Create some configs to merge
        config1 = ConfigIR(
            python="config.foo=6", dataId={"visit": 7}, file=["test1.py"], rest={"a": 1, "b": 2}
        )
        config2 = ConfigIR(python=None, dataId=None, file=["test2.py"], rest={"c": 1, "d": 2})
        config3 = ConfigIR(python="config.bar=7", dataId=None, file=["test3.py"], rest={"c": 1, "d": 2})
        config4 = ConfigIR(python=None, dataId=None, file=["test4.py"], rest={"c": 3, "e": 4})
        config5 = ConfigIR(rest={"f": 5, "g": 6})
        config6 = ConfigIR(rest={"h": 7, "i": 8})
        config7 = ConfigIR(rest={"h": 9})

        # Merge configs with different dataIds, this should yield two elements
        self.assertEqual(list(config1.maybe_merge(config2)), [config1, config2])

        # Merge configs with python blocks defined, this should yield two
        # elements
        self.assertEqual(list(config1.maybe_merge(config3)), [config1, config3])

        # Merge configs with file defined, this should yield two elements
        self.assertEqual(list(config2.maybe_merge(config4)), [config2, config4])

        # merge config2 into config1
        merge_result = list(config5.maybe_merge(config6))
        self.assertEqual(len(merge_result), 1)
        self.assertEqual(config5.rest, {"f": 5, "g": 6, "h": 7, "i": 8})

        # Cant merge configs with shared keys
        self.assertEqual(list(config6.maybe_merge(config7)), [config6, config7])


class PipelineIRTestCase(unittest.TestCase):
    """A test case for PipelineIR objects."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testPipelineIRInitChecks(self):
        # Missing description
        pipeline_str = """
        tasks:
          a: module.A
        """
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Missing tasks
        pipeline_str = """
        description: Test Pipeline
        """
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # This should raise a FileNotFoundError, as there are imported defined
        # so the __init__ method should pass but the imported file does not
        # exist
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports: /dummy_pipeline.yaml
        """
        )

        with self.assertRaises(FileNotFoundError):
            PipelineIR.from_string(pipeline_str)

    def testTaskParsing(self):
        # Should be able to parse a task defined both ways
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test.modA
            modB:
              class: test.modB
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(list(pipeline.tasks.keys()), ["modA", "modB"])
        self.assertEqual([t.klass for t in pipeline.tasks.values()], ["test.modA", "test.modB"])

    def testImportParsing(self):
        # This should raise, as the two pipelines, both define the same label
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline1.yaml
            - $TESTDIR/testPipeline2.yaml
        """
        )
        # "modA" is the duplicated label, and it should appear in the error.
        with self.assertRaisesRegex(ValueError, "modA"):
            PipelineIR.from_string(pipeline_str)

        # This should pass, as the conflicting task is excluded
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline1.yaml
              exclude: modA
            - $TESTDIR/testPipeline2.yaml
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(set(pipeline.tasks.keys()), {"modA", "modB"})

        # This should pass, as the conflicting task is no in includes
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline1.yaml
              include: modB
              labeledSubsetModifyMode: DROP
            - $TESTDIR/testPipeline2.yaml
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(set(pipeline.tasks.keys()), {"modA", "modB"})

        # Test that you cant include and exclude a task
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline1.yaml
              exclude: modA
              include: modB
              labeledSubsetModifyMode: EDIT
            - $TESTDIR/testPipeline2.yaml
        """
        )

        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Test unknown labeledSubsetModifyModes raise
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline1.yaml
              exclude: modA
              include: modB
              labeledSubsetModifyMode: WRONG
            - $TESTDIR/testPipeline2.yaml
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Test that contracts are imported
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline1.yaml
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.contracts[0].contract, "modA.b == modA.c")

        # Test that contracts are not imported
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline1.yaml
              importContracts: False
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.contracts, [])

        # Test that configs are imported when defining the same task again
        # with the same label
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
        tasks:
          modA:
            class: "test.moduleA"
            config:
                value2: 2
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.tasks["modA"].config[0].rest, {"value1": 1, "value2": 2})

        # Test that configs are not imported when redefining the task
        # associated with a label
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
        tasks:
          modA:
            class: "test.moduleAReplace"
            config:
                value2: 2
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.tasks["modA"].config[0].rest, {"value2": 2})

        # Test that named subsets are imported
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.labeled_subsets.keys(), {"modSubset"})
        self.assertEqual(pipeline.labeled_subsets["modSubset"].subset, {"modA"})

        # Test that imported and redeclaring a named subset works
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
        tasks:
          modE: "test.moduleE"
        subsets:
          modSubset:
            - modE
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.labeled_subsets.keys(), {"modSubset"})
        self.assertEqual(pipeline.labeled_subsets["modSubset"].subset, {"modE"})

        # Test that imported from two pipelines that both declare a named
        # subset with the same name fails
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
            - $TESTDIR/testPipeline3.yaml
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Test that imported a named subset that duplicates a label declared
        # in this pipeline fails
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
        tasks:
          modSubset: "test.moduleE"
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Test that imported fails if a named subset and task label conflict
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline2.yaml
            - $TESTDIR/testPipeline4.yaml
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Test that importing Pipelines with different step definitions fails
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - $TESTDIR/testPipeline5.yaml
        steps:
            - label: sub1
              sharding_dimensions: ['a', 'e']
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # Test that it does not fail if steps are excluded
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline5.yaml
              importSteps: false
        steps:
            - label: sub1
              sharding_dimensions: ['a', 'e']
        """
        )
        PipelineIR.from_string(pipeline_str)

        # Test that importing does work
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
            - location: $TESTDIR/testPipeline5.yaml
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(set(step.label for step in pipeline.steps), {"sub1", "sub2"})

    def testSteps(self):
        # Test that steps definitions are created
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: "test.moduleA"
            modB: "test.moduleB"
        subsets:
            sub1:
                subset:
                - modA
                - modB
            sub2:
                subset:
                - modA
        steps:
            - label: sub1
              sharding_dimensions: ['a', 'b']
            - label: sub2
              sharding_dimensions: ['a', 'b']
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(set(step.label for step in pipeline.steps), {"sub1", "sub2"})

        # Test that steps definitions must be unique
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: "test.moduleA"
            modB: "test.moduleB"
        subsets:
            sub1:
                subset:
                - modA
                - modB
            sub2:
                subset:
                - modA
        steps:
            - label: sub1
              sharding_dimensions: ['a', 'b']
            - label: sub1
              sharding_dimensions: ['a', 'b']
        """
        )
        with self.assertRaises(ValueError):
            pipeline = PipelineIR.from_string(pipeline_str)

    def testReadParameters(self):
        # verify that parameters section are read in from a pipeline
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        parameters:
          value1: A
          value2: B
        tasks:
          modA: ModuleA
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.parameters.mapping, {"value1": "A", "value2": "B"})

    def testTaskParameterLabel(self):
        # verify that "parameters" cannot be used as a task label
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
          parameters: modA
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

    def testParameterImporting(self):
        # verify that importing parameters happens correctly
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
          - $TESTDIR/testPipeline1.yaml
          - location: $TESTDIR/testPipeline2.yaml
            exclude:
              - modA

        parameters:
          value4: valued
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(
            pipeline.parameters.mapping,
            {"value4": "valued", "value1": "valueNew", "value2": "valueB", "value3": "valueC"},
        )

    def testImportingInstrument(self):
        # verify an instrument is imported, or ignored, (Or otherwise modified
        # for potential future use)
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
          - $TESTDIR/testPipeline1.yaml
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.instrument, "test.instrument")

        # verify that an imported pipeline can have its instrument set to None
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
          - location: $TESTDIR/testPipeline1.yaml
            instrument: None
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.instrument, None)

        # verify that an imported pipeline can have its instrument modified
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        imports:
          - location: $TESTDIR/testPipeline1.yaml
            instrument: new.instrument
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.instrument, "new.instrument")

        # Test that multiple instruments can't be defined,
        # and that the error message tells you what instruments were found.
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        instrument: new.instrument
        imports:
          - location: $TESTDIR/testPipeline1.yaml
        """
        )
        with self.assertRaisesRegex(ValueError, "new.instrument .* test.instrument."):
            PipelineIR.from_string(pipeline_str)

    def testParameterConfigFormatting(self):
        # verify that a config properly is formatted with parameters
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        parameters:
          value1: A
        tasks:
          modA:
            class: ModuleA
            config:
              testKey: parameters.value1
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        newConfig = pipeline.tasks["modA"].config[0].formatted(pipeline.parameters)
        self.assertEqual(newConfig.rest["testKey"], "A")

    def testReadContracts(self):
        # Verify that contracts are read in from a pipeline
        location = "$TESTDIR/testPipeline1.yaml"
        pipeline = PipelineIR.from_uri(location)
        self.assertEqual(pipeline.contracts[0].contract, "modA.b == modA.c")

        # Verify that a contract message is loaded
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test.modA
            modB:
              class: test.modB
        contracts:
            - contract: modA.foo == modB.Bar
              msg: "Test message"
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.contracts[0].msg, "Test message")

    def testReadNamedSubsets(self):
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test.modA
            modB:
              class: test.modB
            modC: test.modC
            modD: test.modD
        subsets:
            subset1:
              - modA
              - modB
            subset2:
              subset:
                - modC
                - modD
              description: "A test named subset"
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.labeled_subsets.keys(), {"subset1", "subset2"})

        self.assertEqual(pipeline.labeled_subsets["subset1"].subset, {"modA", "modB"})
        self.assertEqual(pipeline.labeled_subsets["subset1"].description, None)

        self.assertEqual(pipeline.labeled_subsets["subset2"].subset, {"modC", "modD"})
        self.assertEqual(pipeline.labeled_subsets["subset2"].description, "A test named subset")

        # verify that forgetting a subset key is an error
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test.modA
            modB:
              class: test.modB
            modC: test.modC
            modD: test.modD
        subsets:
            subset2:
              sub:
                - modC
                - modD
              description: "A test named subset"
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

        # verify putting a label in a named subset that is not in the task is
        # an error
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test.modA
            modB:
              class: test.modB
            modC: test.modC
            modD: test.modD
        subsets:
            subset2:
              - modC
              - modD
              - modE
        """
        )
        with self.assertRaises(ValueError):
            PipelineIR.from_string(pipeline_str)

    def testSubsettingPipeline(self):
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test.modA
            modB:
              class: test.modB
            modC: test.modC
            modD: test.modD
        subsets:
            subset1:
              - modA
              - modB
            subset2:
              subset:
                - modC
                - modD
              description: "A test named subset"
        """
        )
        pipeline = PipelineIR.from_string(pipeline_str)
        # verify that creating a pipeline subset with the default drop behavior
        # removes any labeled subset that contains a label not in the set of
        # all task labels.
        pipelineSubset1 = pipeline.subset_from_labels({"modA", "modB", "modC"})
        self.assertEqual(pipelineSubset1.labeled_subsets.keys(), {"subset1"})
        # verify that creating a pipeline subset with the edit behavior
        # edits any labeled subset that contains a label not in the set of
        # all task labels.
        pipelineSubset2 = pipeline.subset_from_labels({"modA", "modB", "modC"}, PipelineSubsetCtrl.EDIT)
        self.assertEqual(pipelineSubset2.labeled_subsets.keys(), {"subset1", "subset2"})
        self.assertEqual(pipelineSubset2.labeled_subsets["subset2"].subset, {"modC"})

    def testInstrument(self):
        # Verify that if instrument is defined it is parsed out
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        instrument: dummyCam
        tasks:
            modA: test.moduleA
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.instrument, "dummyCam")

    def testReadTaskConfig(self):
        # Verify that a task with a config is read in correctly
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA:
                class: test.moduleA
                config:
                    propertyA: 6
                    propertyB: 7
                    file: testfile.py
                    python: "config.testDict['a'] = 9"
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.tasks["modA"].config[0].file, ["testfile.py"])
        self.assertEqual(pipeline.tasks["modA"].config[0].python, "config.testDict['a'] = 9")
        self.assertEqual(pipeline.tasks["modA"].config[0].rest, {"propertyA": 6, "propertyB": 7})

        # Verify that multiple files are read fine
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA:
                class: test.moduleA
                config:
                    file:
                        - testfile.py
                        - otherFile.py
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.tasks["modA"].config[0].file, ["testfile.py", "otherFile.py"])

        # Test reading multiple Config entries
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA:
                class: test.moduleA
                config:
                    - propertyA: 6
                      propertyB: 7
                      dataId: {"visit": 6}
                    - propertyA: 8
                      propertyB: 9
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)
        self.assertEqual(pipeline.tasks["modA"].config[0].rest, {"propertyA": 6, "propertyB": 7})
        self.assertEqual(pipeline.tasks["modA"].config[0].dataId, {"visit": 6})
        self.assertEqual(pipeline.tasks["modA"].config[1].rest, {"propertyA": 8, "propertyB": 9})
        self.assertEqual(pipeline.tasks["modA"].config[1].dataId, None)

    def testSerialization(self):
        # Test creating a pipeline, writing it to a file, reading the file
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        instrument: dummyCam
        imports:
          - location: $TESTDIR/testPipeline1.yaml
            instrument: None
        tasks:
            modC:
                class: test.moduleC
                config:
                    - propertyA: 6
                      propertyB: 7
                      dataId: {"visit": 6}
                    - propertyA: 8
                      propertyB: 9
            modD: test.moduleD
        contracts:
            - modA.foo == modB.bar
        subsets:
            subA:
               - modA
               - modC
        """
        )

        pipeline = PipelineIR.from_string(pipeline_str)

        # Create the temp file, write and read
        with tempfile.NamedTemporaryFile() as tf:
            pipeline.write_to_uri(tf.name)
            loaded_pipeline = PipelineIR.from_uri(tf.name)
        self.assertEqual(pipeline, loaded_pipeline)

    def testPipelineYamlLoader(self):
        # Tests that an exception is thrown in the case a key is used multiple
        # times in a given scope within a pipeline file
        pipeline_str = textwrap.dedent(
            """
        description: Test Pipeline
        tasks:
            modA: test1
            modB: test2
            modA: test3
        """
        )
        self.assertRaises(KeyError, PipelineIR.from_string, pipeline_str)

    def testMultiLineStrings(self):
        """Test that multi-line strings in pipelines are written with
        '|' continuation-syntax instead of explicit newlines.
        """
        pipeline_ir = PipelineIR({"description": "Line 1\nLine2\n", "tasks": {"modA": "task1"}})
        string = str(pipeline_ir)
        self.assertIn("|", string)
        self.assertNotIn(r"\n", string)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from lsst.pipe.base.tests.mocks import DataIdMatch


class DataIdMatchTestCase(unittest.TestCase):
    """A test case for DataidMatch class."""

    dataIds = (
        {"instrument": "INSTR", "detector": 1, "number": 4},
        {"instrument": "INSTR", "detector": 2, "number": 3},
        {"instrument": "LSST", "detector": 3, "number": 2},
        {"instrument": "LSST", "detector": 4, "number": 1},
    )

    def test_strings(self):
        """Tests for string comparisons method."""
        tests = (
            ("instrument = 'INSTR'", [True, True, False, False]),
            ("instrument = 'LSST'", [False, False, True, True]),
            ("instrument < 'LSST'", [True, True, False, False]),
            ("instrument IN ('LSST', 'INSTR')", [True, True, True, True]),
        )

        for expr, result in tests:
            dataIdMatch = DataIdMatch(expr)
            self.assertEqual([dataIdMatch.match(dataId) for dataId in self.dataIds], result)

    def test_comparisons(self):
        """Test all supported comparison operators."""
        tests = (
            ("detector = 1", [True, False, False, False]),
            ("detector != 1", [False, True, True, True]),
            ("detector > 2", [False, False, True, True]),
            ("2 <= detector", [False, True, True, True]),
            ("2 > detector", [True, False, False, False]),
            ("2 >= detector", [True, True, False, False]),
        )

        for expr, result in tests:
            dataIdMatch = DataIdMatch(expr)
            self.assertEqual([dataIdMatch.match(dataId) for dataId in self.dataIds], result)

    def test_arith(self):
        """Test all supported arithmetical operators."""
        tests = (
            ("detector + number = 5", [True, True, True, True]),
            ("detector - number = 1", [False, False, True, False]),
            ("detector * number = 6", [False, True, True, False]),
            ("detector / number = 1.5", [False, False, True, False]),
            ("detector % number = 1", [True, False, True, False]),
            ("+detector = 1", [True, False, False, False]),
            ("-detector = -4", [False, False, False, True]),
        )

        for expr, result in tests:
            dataIdMatch = DataIdMatch(expr)
            self.assertEqual([dataIdMatch.match(dataId) for dataId in self.dataIds], result)

    def test_logical(self):
        """Test all supported logical operators."""
        tests = (
            ("detector = 1 OR instrument = 'LSST'", [True, False, True, True]),
            ("detector = 1 AND instrument = 'INSTR'", [True, False, False, False]),
            ("NOT detector = 1", [False, True, True, True]),
        )

        for expr, result in tests:
            dataIdMatch = DataIdMatch(expr)
            self.assertEqual([dataIdMatch.match(dataId) for dataId in self.dataIds], result)

    def test_parens(self):
        """Test parentheses."""
        tests = (("(detector = 1 OR number = 1) AND instrument = 'LSST'", [False, False, False, True]),)

        for expr, result in tests:
            dataIdMatch = DataIdMatch(expr)
            self.assertEqual([dataIdMatch.match(dataId) for dataId in self.dataIds], result)

    def test_in(self):
        """Test IN expression."""
        tests = (
            ("detector in (1, 3, 2)", [True, True, True, False]),
            ("detector not in (1, 3, 2)", [False, False, False, True]),
            ("detector in (1..4:2)", [True, False, True, False]),
            ("detector in (1..4:2, 4)", [True, False, True, True]),
        )

        for expr, result in tests:
            dataIdMatch = DataIdMatch(expr)
            self.assertEqual([dataIdMatch.match(dataId) for dataId in self.dataIds], result)

    def test_errors(self):
        """Test for errors in expressions."""
        dataId = {"instrument": "INSTR", "detector": 1}

        # Unknown identifier
        expr = "INSTRUMENT = 'INSTR'"
        dataIdMatch = DataIdMatch(expr)
        with self.assertRaisesRegex(KeyError, "INSTRUMENT"):
            dataIdMatch.match(dataId)

        # non-boolean expression
        expr = "instrument"
        dataIdMatch = DataIdMatch(expr)
        with self.assertRaisesRegex(TypeError, "Expression 'instrument' returned non-boolean object"):
            dataIdMatch.match(dataId)

        # operations on unsupported combination of types
        expr = "instrument - detector = 0"
        dataIdMatch = DataIdMatch(expr)
        with self.assertRaisesRegex(TypeError, "unsupported operand type"):
            dataIdMatch.match(dataId)

        # function calls are not implemented
        expr = "POINT(2, 1) != POINT(1, 2)"
        dataIdMatch = DataIdMatch(expr)
        with self.assertRaises(NotImplementedError):
            dataIdMatch.match(dataId)


if __name__ == "__main__":
    unittest.main()
